/**
 * Header can either be,
 * 1. An object
 * 2. A function that return an object or undefined
 * 3. A function that returns a Promise that resolves to an object or undefined
 *
 * The function types can return undefined instead of an object because there are cases where
 * we still want the API call to run even if the header cannot be generated. Such as if we were
 * to use a function to read for a auth token in local storage and only return the object with
 * the token if it exists. This will not cause any errors because spreading `undefined` into an
 * object will just do nothing { something: true, ...undefined } will just be { something: true }
 *
 * Header cannot be `undefined` directly because the `header` method cannot be directly called
 * with `undefined` because that wouldn't make any sense.
 *
 * If function throws, the API call will be cancelled and the error will be caught by the 'safe'
 * functions i.e. the 'run' methods to be returned to users.
 *
 * What if the API call should be cancelled if there is no header generated? Should it be on the
 * library to quit when `undefined` is returned or should it be on the users to throw an error if
 * no headers can be generated? Might be leaning more to library handling so that user land code
 * does not have to repetitively implement throws.
 * This might even tie in to the plugin idea where a plugin can cancel an API call like middlewares.
 *
 * Exporting this type so that you can explicitly type your Header objects
 * with this to ensure that it is correctly typed at point of value definition
 * instead of only type checking when you call the `.header` method.
 */
export declare type Header = Record<string, any> | (() => Record<string, any> | undefined) | (() => Promise<Record<string, any> | undefined>);
/**
 * All the supported HTTP methods.
 * Ref: https://fetch.spec.whatwg.org/#methods
 *
 * Based on the reference link, although there are other HTTP methods, they are not
 * supported by the `fetch` spec, and therefore not included in this union type as
 * even if you could pass it to the `fetch` function, it will just fail.
 *
 * `HEAD` and `OPTIONS` HTTP methods are quite special and low level and extremely
 * rarely used, therefore these are the only 2 methods which do not have static
 * constructor wrapper methods to easily create `oof` instances for these methods.
 * Because of that, the only times where a library user will use the constructor
 * directly is when they want to use one of these 2 HTTP methods.
 *
 * Exporting this type so that you can explicitly type your HTTP method strings
 * as needed to ensure that they are correctly typed at point of value definition
 * instead of only type checking when you call the constructor.
 */
export declare type HTTPMethod = "GET" | "POST" | "PUT" | "PATCH" | "DELETE" | "HEAD" | "OPTIONS";
/**
 * JsonResponse type is used as the base type for what can be returned from the server when using `runJSON()`.
 *
 * Note that this type is not a JSON type as this only support `{}` based types and all the other JSON string forms are not supported such as arrays or single strings.
 *
 * The reason why the other types are not supported is because the runJSON method actually injects the values of `ok` and `status` into the response before returning it.
 *
 * Exporting this type so that you can explicitly type your body objects
 * with this to ensure that it is correctly typed at point of value definition
 * instead of only type checking when you call the `.runJSON` method.
 */
export declare type JsonResponse = Record<string | number | symbol, any>;
/**
 * oof: Object Oriented `Fetch` abstraction
 *
 * This object oriented approach gives users a easy to use chainable interface to build their API calls
 */
export declare class oof {
    #private;
    /** Static function to read the value of the base URL */
    static baseUrl: () => string;
    /**
     * Set base url and get back `oof` to chain this method call.
     *
     * Although any static methods can be chained with this method, usually the only thing that
     * you should chain to this method call is the `defaultOptions` method because these 2 methods
     * are usually all that every users need to use in order to setup a base configuration for all
     * future API calls made with this library.
     *
     * And the good thing about chaining with `defaultOptions` static method is that it does not
     * return the `oof` class which means you cannot chain anymore methods to it, and this is good
     * because your initial library configuration call should only consist of setting the `baseUrl`
     * and the default `RequestInit` options.
     */
    static setBaseURL(url: string): typeof oof;
    /**
     * Static method to set default options that will be applied to all API calls.
     * The options set here can be overwritten one-off in specific API calls using
     * the `options` method.
     *
     * Note that everytime this static method is called, all the default options
     * are overwritten to use this, the default options will not be merged.
     *
     * This static method does not return the `oof` class to specifically be
     * unchainable, so that users do not mistakenly use this in an API call,
     * and it will be more explicit when they set default options.
     */
    static defaultOptions(opts: RequestInit): void;
    /**
     * Low level constructor API that generally isnt used.
     *
     * ### When should you use the constructor directly?
     * For most library users, just stick with the provided static methods for a cleaner API.
     *
     * The only times where a library user should use the constructor directly is when they
     * want to use either the `HEAD` or `OPTIONS` HTTP method because they are quite special
     * and low level and extremely rarely used, therefore these are the only 2 methods which
     * do not have static constructor wrapper methods to easily create `oof` instances.
     *
     * ### Parameters
     * The 2 parameters, `method` and `string`, are the only parameters that truly needs to be
     * defined for every single API call and cannot be optional. All other configuration options
     * can be set using instance methods like `options` and `headers` without relying on the
     * constructor to set all of them.
     *
     * ### Compared to v7 and earlier
     * In previous versions of this library, the constructor supported initializing all the
     * options used in the `fetch` function. Although more flexible, it did not make sense
     * because if someone were to use the constructor directly and pass in all the options,
     * then they might as well just use `fetch` directly as that is easier to use when passing
     * in all the config options at once. The whole purpose of this library is to configure
     * the API call parameters using a chainable interface before making that API call, so
     * there is no point in using a constructor that can initialize all of those values.
     * Therefore, in version 8, the constructor API is simplified to remove all the unnecessary
     * parameters to ensure users use this in the way it is designed. This new API also have
     * some added benefits of being smaller in size for the constructor code AND also simplifies
     * the caller code, while making it faster by removing the need to destructure out values,
     * and by skipping the check to possibly transform the headers parameter.
     *
     * ### v7 of the constructor
     * ```typescript
     * constructor({
     *     method,
     *     path,
     *     opts = {},
     *     headers = [],
     *   }: {
     *     method: HTTPMethod;
     *     path: string;
     *     opts?: RequestInit;
     *     headers?: Header | Array<Header>;
     *   }) {
     *     this.#method = method;
     *     this.#path = path;
     *     this.#opts = opts;
     *
     *     // Ensures that this.#headers is always an array regardless of what the user passes in
     *     // Users can pass in a single header object/function or an array of header objects/functions
     *     // If an array is passed in, leave it as it is, else wrap the single header object/function in a array
     *     this.#headers = Array.isArray(headers) ? headers : [headers];
     *   }
     * ```
     */
    constructor(method: HTTPMethod, path: string);
    /**
     * Wrapper function over constructor to construct a new `oof` instance for a `GET` API call
     *
     * @returns {oof} Returns a new instance of `oof` after constructing it to let you chain method calls
     */
    static GET: (path: string) => oof;
    /**
     * Wrapper function over constructor to construct a new `oof` instance for a `POST` API call
     *
     * @returns {oof} Returns a new instance of `oof` after constructing it to let you chain method calls
     */
    static POST: (path: string) => oof;
    /**
     * Wrapper function over constructor to construct a new `oof` instance for a `PUT` API call
     *
     * @returns {oof} Returns a new instance of `oof` after constructing it to let you chain method calls
     */
    static PUT: (path: string) => oof;
    /**
     * Wrapper function over constructor to construct a new `oof` instance for a `PATCH` API call
     *
     * See this link on the difference between POST, PUT and PATCH HTTP methods:
     * https://en.wikipedia.org/wiki/PATCH_(HTTP)#:~:text=The%20main%20difference%20between%20the,instructions%20to%20modify%20the%20resource.
     *
     * @returns {oof} Returns a new instance of `oof` after constructing it to let you chain method calls
     */
    static PATCH: (path: string) => oof;
    /**
     * Wrapper function over constructor to construct a new `oof` instance for a `DEL` API call
     *
     * @returns {oof} Returns a new instance of `oof` after constructing it to let you chain method calls
     */
    static DEL: (path: string) => oof;
    /**
     * Call this method to make a once off API call, that is, to make the API call without
     * using the `baseUrl` of this class. Use this when you need to make an API call to
     * another domain/baseUrl without changing the `baseUrl` for all other API calls.
     *
     * This method sets the `#once` flag on this `oof` instance so that `_run` will treat
     * the URL request path as a full path rather than a subpath to be concatenated to the
     * `#baseUrl`. The flag setting way is used instead of relying on heuristics in the
     * `run` method to detect if the URL is a full path or not.
     *
     * @returns {oof} Returns the current instance of `oof` to let you chain method calls
     */
    once(): oof;
    /**
     * @param {RequestInit} opts RequestInit object for this one API call
     * @returns {oof} Returns the current instance of `oof` to let you chain method calls
     *
     * ### When to use this method?
     * Use this to set custom RequestInit parameters for this specific `oof` instance's
     * fetch method call. See below on when should you use this method.
     *
     * Note that the options set using this method cannot override the headers set with
     * the `headers` method, cannot override the HTTP `method` set in the constructor
     * and cannot override the body value set by any of the 'body' methods.
     *
     * Which is why:
     * - If you want to set request body, use the `.body` or `.bodyJSON` method instead.
     * - If you need to set a request header, use the `.header` method instead.
     * - If you want to set HTTP request method, use one of the static constructors
     * like `oof.GET("/api/path")` or use the constructor by passing in a non-default
     * (not GET/POST/PUT/DEL) HTTP method instead like `new oof({ method:"HTTP-METHOD" })`
     *
     * Do not use this unless you have a specific option to pass in e.g. cache: "no-cache"
     * for this one specific API call only and nothing else. If there are options that you
     * want to set for all API request calls, use the `oof.defaultOptions` static method
     * instead. Note that any options value set using this method will also override the
     * default options set using `oof.defaultOptions`.
     *
     * This method directly assigns the arguement to `this.#opts` which means calling this
     * method overrides whatever options that is already set previously. Because it does
     * not make sense for the user to call this repeatedly since there is no default options
     * set by this library anyways. Thus it is a direct assignment instead of a merge like
     * `this.#opts = { ...this.#opts, ...opts }`
     */
    options(opts: RequestInit): oof;
    /**
     * Add Header(s) to include in the API call.
     *
     * Accepts plain header objects, functions and async functions.
     *
     * Functions passed in will be called right before the API call to generate a header object,
     * to delay generating certain header values like a time limited auth token or recaptcha.
     *
     * This method can be called multiple times, and all the header objects will be combined.
     * If there are duplicate headers, the latter will be used.
     *
     * If you prefer, this method is also a variadic function that accepts multiple arguments
     * of the same type so that you can call this method just once if you have all the headers
     * instead of invoking this method multiple times.
     *
     * @returns {oof} Returns the current instance of `oof` to let you chain method calls
     */
    header(...header: Header[]): oof;
    /**
     * Set the request body to be sent to server for HTTP methods such as POST/PUT.
     *
     * ### When to use this method?
     * For most users who want to send JSON data to the server, see the `bodyJSON` method
     * instead of a simpler API. For other types of data like `FormData`, `Blob` and `streams`
     * can just pass it into this method as the `body` parameter and the content type will be
     * automatically detected / set by the `fetch` function.
     *
     * ### Why have both `body` and `bodyJSON` method?
     * The reason for this method instead of just having `bodyJSON` only is because the library
     * cannot always just assume that users only use JSON data, and have to support data types
     * like FormData, Blob and etc... However the problem is that when content-type cannot be
     * set fetch will try to guess it, but when body is the results of JSON.stringify(this.#body),
     * fetch will guess the content-type to be text/plain and the browser will treat it as a safe
     * CORS request, which means that for that request there will be no pre-flight request sent.
     * Which means that the browser prevents certain headers from being used, which might cause
     * an issue, and also the server may not always respond correctly because they assume they got
     * text/plain even though your API endpoint is for application/json.
     *
     * The type of `body` value can be anything, as you can pass in any value that the
     * `fetch` API's `RequestInit`'s body property accepts.
     *
     * ### Using generics for TS Type Safety
     * ```javascript
     * const { res, err } = await oof
     *   .POST("/api")
     *   .body<FormData>(someValue) // TS will enforce that someValue must be FormData
     *   .run();
     * ```
     * The above code sets the body type for type safety.
     *
     * For TS users, the body type is a generic type variable even though its default type
     * for it is `any` so that you can use it to restrict the type passed into the method.
     * This allows you to enforce type safety where once a generic type is set, you know that
     * the value passed in for the `body` parameter cannot be any other type.
     *
     * ### On `optionalContentType`'s type safety
     * Note on `optionalContentType`'s type: Although it is possible to create a union type
     * of all allowed string literals for the content-type header / mime types, it is not
     * very feasible as it is a very big list that will be updated in the future. Therefore
     * users are expected to make sure that any string they pass is valid.
     * References on all supported content-type values:
     * - https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Content-Type
     * - https://stackoverflow.com/questions/23714383/what-are-all-the-possible-values-for-http-content-type-header/48704300#48704300
     * - https://www.iana.org/assignments/media-types/media-types.xhtml
     *
     * @returns {oof} Returns the current instance of `oof` to let you chain method calls
     */
    body<T = any>(body: T, optionalContentType?: string): oof;
    /**
     * @param data Any data type that is of 'application/json' type and can be stringified by JSON.stringify
     * @returns {oof} Returns the current instance of `oof` to let you chain method calls
     *
     * ### About
     * Method that stringifies a JSON stringifiable data type to use as the request body,
     * and sets the content-type to 'application/json'.
     *
     * ### What data type can be passed in?
     * The type of data that can be passed in, is any JS value that can be JSON serialized
     * with `JSON.stringify()`. See this [MDN link](https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/JSON/stringify#description)
     * on what type of data can be passed in.
     *
     * ### What if you have no data to send?
     * Even though there is no default arguement, you do not have to call the `.data` method with an empty object when
     * calling a method like `oof.POST` as `fetch` and API services will just treat it as an empty object by default.
     *
     * ### Why does this method exists?
     * Because, the `_run` method needs to accept too many data input types,
     * so instead of doing all the processing and transforming then, the specific
     * transformations should be done in helper methods like this so that in the
     * actual `fetch` call, we can just set `body: this.#body` directly.
     *
     * ### Why not just use the `body` method?
     * Since JSON is one of the most popular methods of communication over HTTP,
     * this method helps users to write less code by helping them stringify their
     * JS object and set the content-type header to 'application/json', instead
     * of requiring them to explicitly call `.body(JSON.stringify(data), "application/json")`
     * every single time they want to send JSON data to their API servers.
     *
     * ### Using generics for TS Type Safety
     * ```javascript
     * const { res, err } = await oof
     *   .POST("/api")
     *   .bodyJSON<MyRequestBody>(val) // TS will enforce that val must be MyRequestBody
     *   .run();
     * ```
     * The above code sets the body type for type safety.
     *
     * For TS users, the data parameter have a generic type even though its default type
     * is `any` so that you can use it to restrict the type passed into the method.
     * This allows you to enforce type safety where once a generic type is set, you know that
     * the value passed in for the `body` parameter cannot be any other type.
     */
    bodyJSON<T = any>(data: T): oof;
    /**
     * ### About
     * This is the underlying raw `_run` method, that is called internally after constructing
     * the API call object to make the API call, and it should not be used by users directly.
     * See the other run methods, `run`, `runJSON`, `runText`, `runBlob`, `runFormData`, `runArrayBuffer`.
     *
     * ### Method 'safety'
     * All other 'run' methods are safe by default, i.e. they do not throw on any errors/exceptions!
     * This is the only underlying raw method that might throw an error/exception when something goes
     * wrong. All the other methods are wrapped in the `safe` function to catch any errors so that it
     * can be returned instead of causing a jump in the code control flow to the nearest catch block.
     *
     * The safety feature is super useful as it reduces the amount of boiler plate code you have to
     * write (try/catch blocks and .catch methods) when dealing with libraries that can throw as it
     * will disrupt your own code's flow. This safe APIs enables you to write single block level code
     * that are guaranteed to not throw and gives you a super readable code control flow!
     *
     * ### What is this method about?
     * This method is basically a wrapper around the fetch API. After configuring all the values using
     * the object oriented notation (method chaining), when you call `_run`, it basically takes all the
     * values on its instance props and use these as option values for the fetch API's RequestInit
     * parameter while taking care of certain things like creating the full API url using any baseUrl
     * set with `setBaseUrl`, delayed header generation and etc...
     *
     * ### More on the return type
     * The return type is unioned with `never` because this function can throw, aka never return.
     * However with TS, any value unioned with `never`, will be itself, because checking for control
     * flow is not enforced / not possible with TS. This is more for documentation purposes for
     * library writers and users who want to learn more about this API.
     */
    _run(): Promise<Response> | never;
    /**
     * Safe version of the `_run` method that **will not throw** or let any errors bubble up,
     * i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.GET("/api").run();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be the Response object
     * ```
     */
    run(): Promise<{
        res: undefined;
        err: Error;
    } | {
        res: Response;
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `_run` method to return response body parsed as text.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.GET("/api").runText();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be be a string
     * ```
     */
    runText(): Promise<{
        res: undefined;
        err: Error;
    } | {
        res: string;
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `_run` method to return response body parsed as Blob.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.GET("/api").runBlob();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be a Blob
     * ```
     */
    runBlob(): Promise<{
        res: undefined;
        err: Error;
    } | {
        res: Blob;
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `_run` method to return response body parsed as FormData.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.GET("/api").runFormData();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be form data
     * ```
     */
    runFormData(): Promise<{
        res: undefined;
        err: Error;
    } | {
        res: FormData;
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `_run` method to return response body parsed as ArrayBuffer.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.GET("/api").runArrayBuffer();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be an Array Buffer
     * ```
     */
    runArrayBuffer(): Promise<{
        res: undefined;
        err: Error;
    } | {
        res: ArrayBuffer;
        err: undefined;
    }>;
    /**
     * Abstraction on top of the `_run` method to return response body parsed as JSON.
     *
     * This method is 'safe' in the sense that this **will not throw** or let any errors bubble
     * up, i.e. no try/catch or .catch method needed to handle the jumping control flow of errors.
     *
     * @example <caption>Call API and handle any errors sequentially at the same scope level</caption>
     * ```javascript
     * const { res, err } = await oof.GET("/api").runJSON<MyResponseObjectType>();
     *
     * if (err) return console.log("API Call failed!");
     *
     * console.log("Res:", res); // Type narrowed to be 'MyResponseObjectType'
     * ```
     *
     * ### Why are values injected in the return type?
     * Return type will always union with { ok: boolean; status: number; } as these will always be injected in.
     *
     * When API server responds with a status code of anything outside of 200-299 Response.ok is auto set to false
     * https://developer.mozilla.org/en-US/docs/Web/API/Fetch_API/Using_Fetch#checking_that_the_fetch_was_successful
     * https://developer.mozilla.org/en-US/docs/Web/API/Response/ok
     *
     * Thus instead of making API servers include an 'ok' data prop in response body,
     * this method auto injects in the ok prop using Response.ok as long as API server use the right HTTP code.
     * However the 'ok' prop is set before the spread operator so your API can return an 'ok' to override this.
     *
     * ### What about redirects?
     * You do not have to worry about this method setting `ok` to false when the HTTP response code is 300-399
     * even though it is outside of the 2XX range because by default, fetch API's option will have redirect
     * set to "follow", which means it will follow the redirect to the final API end point and only then is
     * the `ok` value set with the final HTTP response code.
     *
     * ### Using generics for TS Type Safety
     * For TS users, this method accepts a generic type to type the returned object. Allowing you to have type
     * safety for the response object. However, this DOES NOT perform any runtime data validation, so even if
     * it is type safe, it does not mean that the response object is guaranteed to be what you typed it to be.
     *
     * @todo
     * Might include a way to do response object validation, but TBD on how to implement it without bloating the library.
     *
     * Record is keyed by any type `string|number|Symbol` which an object can be indexed with
     */
    runJSON<T extends JsonResponse = JsonResponse>(): Promise<{
        res: undefined;
        err: Error;
    } | {
        res: T & {
            ok: boolean;
            status: number;
        };
        err: undefined;
    }>;
}
//# sourceMappingURL=index.d.ts.map